// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package hints

import (
	"encoding/json"
	"strings"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/utils"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"

	"go.uber.org/zap/zapcore"
)

type HintRequestMessage struct {
	OrgKey         string `json:"org_key" binding:"required"`
	DeviceGroupId  int64  `json:"device_group_id" binding:"min=1"`
	DeviceId       int64  `json:"device_id" binding:"min=1"`
	DeviceType     devicetype.DeviceType
	DeploymentType deploymenttype.DeploymentType
	HintRequestBase
	AuditHintResult *HintResult
}

type AuditHintsFromCBD struct {
	AuditHints                  []string `json:"audit_hints,omitempty"`
	AuditLastUpdateTimestamp    *int64   `json:"audit_last_update_time,omitempty"`
	AuditLastLiveQueryTimestamp *int64   `json:"audit_last_live_query_time,omitempty"`
}

type HintRequestBase struct {
	Version             int
	LastUpdateTime      *time.Time
	LastLiveQueryTime   *time.Time
	RemoteRulesRequired bool
}

func (hrb HintRequestBase) MarshalLogObject(encoder zapcore.ObjectEncoder) error {
	encoder.AddInt("request_version", hrb.Version)
	utils.AddTimePointerToEncoder(encoder, "request_last_update_time", hrb.LastUpdateTime)
	utils.AddTimePointerToEncoder(encoder, "request_last_live_query_time", hrb.LastLiveQueryTime)
	encoder.AddBool("request_remote_rules_required", hrb.RemoteRulesRequired)
	return nil
}

type JsonHintRequest struct {
	// Unique organization identifier within an environment.
	//
	// required: true
	// example: 7ZP4H5XR
	OrgKey string `json:"org_key" binding:"required"`

	// Unique device group identifier.
	//
	// required: true
	// example: 456
	DeviceGroupId int64 `json:"device_group_id" binding:"min=1"`

	// Unique device identifier.
	//
	// required: true
	// example: 123
	DeviceId int64 `json:"device_id" binding:"min=1"`

	// The device's type; if not specified, will have a value of "UNKNOWN".
	//
	// Valid Device Types:
	// 	ANDROID,
	//	IOS,
	//	LINUX,
	//	MAC,
	//	OTHER,
	//	WINDOWS
	//
	// required: false
	// example: WINDOWS
	DeviceType string `json:"device_type"`

	// The device's deployment type; if not specified, will have a value of "UNKNOWN".
	//
	// Valid Deployment Types:
	//	AWS,
	//	AZURE,
	//	ENDPOINT,
	//	GCP,
	//	VDI,
	//	WORKLOAD
	//
	// required: false
	// example: ENDPOINT
	DeploymentType string `json:"deployment_type"`

	// The CBC Protobuf version.
	//
	// required: false
	// example: 18
	Version *int `json:"version"`

	// The timestamp when the sensor last received hints.
	//
	// required: false
	// example: 1672531200000
	LastUpdateTime *int64 `json:"last_update_time"`

	// The timestamp when the sensor was last instructed to get live query commands.
	//
	// required: false
	// example: 1672531200000
	LastLiveQueryTime *int64 `json:"last_live_query_time"`

	// Flag if remote rules required; if True and not a REQ_HELLO message code, no hints will be generated.
	//
	// required: false
	// default: false
	RemoteRulesRequired bool `json:"remote_rules_required"`

	// swagger:ignore
	AuditHintsFromCBD
}

func NewHintRequestBase(version int, lastUpdateTime, lastLiveQueryTime *time.Time, remoteRulesRequired bool) HintRequestBase {
	return HintRequestBase{Version: version, LastUpdateTime: lastUpdateTime, LastLiveQueryTime: lastLiveQueryTime, RemoteRulesRequired: remoteRulesRequired}
}

func NewAuditHintsFromCBD(auditHints []string, auditLastUpdateTimestamp, auditLivequeryTimestamp *int64) AuditHintsFromCBD {
	return AuditHintsFromCBD{
		AuditHints:                  auditHints,
		AuditLastUpdateTimestamp:    auditLastUpdateTimestamp,
		AuditLastLiveQueryTimestamp: auditLivequeryTimestamp,
	}
}

type HintRequest struct {
	Device      *models.Device
	org         *models.Organization
	environment *models.Environment
	deviceGroup *models.DeviceGroup
	HintRequestBase
}

func TimePtrToInt64Ptr(theTime *time.Time) *int64 {
	if theTime == nil {
		return nil
	}

	millis := theTime.UnixMilli()
	return &millis
}

func (hrm *HintRequestMessage) AuditHintsFromRequest() (auditInfo *AuditHintsFromCBD) {
	if hrm.AuditHintResult == nil {
		return
	}

	auditInfo = &AuditHintsFromCBD{
		AuditLastUpdateTimestamp:    TimePtrToInt64Ptr(hrm.AuditHintResult.LastUpdateTime),
		AuditLastLiveQueryTimestamp: TimePtrToInt64Ptr(hrm.AuditHintResult.LastLiveQueryTime),
	}
	if hrm.AuditHintResult.Hints != nil {
		auditInfo.AuditHints = hrm.AuditHintResult.Hints.ToStringArray()
	}

	return
}

// UnmarshalJSON -- Overlay a JSON representation onto an existing HintRequestMessage structure
func (hrm *HintRequestMessage) UnmarshalJSON(b []byte) error {
	jhr := new(JsonHintRequest)
	err := json.Unmarshal(b, jhr)
	if err != nil {
		return err
	}

	hrm.OrgKey = strings.ToUpper(jhr.OrgKey) // Sanitize casing
	hrm.DeviceGroupId = jhr.DeviceGroupId
	hrm.DeviceId = jhr.DeviceId
	hrm.DeviceType = devicetype.FromString(jhr.DeviceType)
	hrm.DeploymentType = deploymenttype.FromString(jhr.DeploymentType)
	if jhr.Version != nil {
		hrm.Version = *jhr.Version
	} else {
		hrm.Version = BASEMESSAGE.PSC_VERSION
	}

	if jhr.LastUpdateTime != nil {
		foundLastUpdateTime := time.UnixMilli(*jhr.LastUpdateTime)
		hrm.LastUpdateTime = &foundLastUpdateTime
	}

	if jhr.LastLiveQueryTime != nil {
		foundLiveQueryTime := time.UnixMilli(*jhr.LastLiveQueryTime)
		hrm.LastLiveQueryTime = &foundLiveQueryTime
	}

	hrm.RemoteRulesRequired = jhr.RemoteRulesRequired

	// check if these are found in jhr, only then use.
	if jhr.AuditHints != nil {
		auditHintResult := new(HintResult)
		var expectedHints hinttype.HintSet
		expectedHints.ConvertFromHintStringArray(jhr.AuditHints)
		auditHintResult.Hints = &expectedHints
		if jhr.AuditLastUpdateTimestamp != nil {
			auditHintResult.LastUpdateTime = utils.GetTimePointer(*jhr.AuditLastUpdateTimestamp)
		}
		if jhr.AuditLastLiveQueryTimestamp != nil {
			auditHintResult.LastLiveQueryTime = utils.GetTimePointer(*jhr.AuditLastLiveQueryTimestamp)
		}
		hrm.AuditHintResult = auditHintResult
	}

	return nil
}

func NewHintRequestFromBase(
	environment *models.Environment,
	org *models.Organization,
	deviceGroup *models.DeviceGroup,
	device *models.Device,
	base HintRequestBase,
) *HintRequest {
	return &HintRequest{environment: environment, org: org, deviceGroup: deviceGroup, Device: device, HintRequestBase: base}
}

func NewHintRequest(environment *models.Environment,
	org *models.Organization,
	deviceGroup *models.DeviceGroup,
	device *models.Device,
	version int, lastUpdateTime *time.Time,
	lastLiveQueryTime *time.Time,
	remoteRulesRequired bool,
) *HintRequest {
	return &HintRequest{environment: environment, org: org, deviceGroup: deviceGroup, Device: device, HintRequestBase: HintRequestBase{Version: version, LastUpdateTime: lastUpdateTime, LastLiveQueryTime: lastLiveQueryTime, RemoteRulesRequired: remoteRulesRequired}}
}

// NewHintRequestMessage -- Create a new hint request for testing
//   - DeviceType and DeploymentType not set, add to returned value as needed
func NewHintRequestMessage(orgKey string, deviceGroupID, deviceID int64,
	version int, lastUpdateTime *time.Time, lastLiveQueryTime *time.Time, remoteRulesRequired bool, deviceType devicetype.DeviceType, deploymentType deploymenttype.DeploymentType,
) *HintRequestMessage {
	return &HintRequestMessage{
		DeviceType: deviceType, DeploymentType: deploymentType, OrgKey: orgKey, DeviceGroupId: deviceGroupID, DeviceId: deviceID, HintRequestBase: NewHintRequestBase(version, lastUpdateTime, lastLiveQueryTime, remoteRulesRequired),
	}
}

// NewHintRequestMessageWithAuditHints -- Create a new hint request for testing with audit information
//   - DeviceType and DeploymentType not set, add to returned value as needed
func NewHintRequestMessageWithAuditHints(orgKey string, deviceGroupID, deviceId int64,
	version int, lastUpdateTime *time.Time, lastLiveQueryTime *time.Time, remoteRulesRequired bool, deviceType devicetype.DeviceType, deploymentType deploymenttype.DeploymentType, hintsFromCBD AuditHintsFromCBD,
) *HintRequestMessage {
	auditHintResult := new(HintResult)
	if hintsFromCBD.AuditHints != nil {
		var expectedHints hinttype.HintSet
		expectedHints.ConvertFromHintStringArray(hintsFromCBD.AuditHints)
		auditHintResult.Hints = &expectedHints
		if hintsFromCBD.AuditLastUpdateTimestamp != nil {
			auditHintResult.LastUpdateTime = utils.GetTimePointer(*hintsFromCBD.AuditLastUpdateTimestamp)
		}
		if hintsFromCBD.AuditLastLiveQueryTimestamp != nil {
			auditHintResult.LastLiveQueryTime = utils.GetTimePointer(*hintsFromCBD.AuditLastLiveQueryTimestamp)
		}
	}
	hrm := NewHintRequestMessage(orgKey, deviceGroupID, deviceId, version, lastUpdateTime, lastLiveQueryTime, remoteRulesRequired, deviceType, deploymentType)
	hrm.AuditHintResult = auditHintResult
	return hrm
}

func (hrm *HintRequestMessage) ToJsonMessage() *JsonHintRequest {
	auditHintsFromCBD := hrm.AuditHintsFromRequest()

	var lastUpdate *int64 = nil
	if hrm.LastUpdateTime != nil {
		lastUpdateTime := hrm.LastUpdateTime.UnixMilli()
		lastUpdate = &lastUpdateTime
	}
	var lastLiveQueryTime *int64 = nil
	if hrm.LastLiveQueryTime != nil {
		lastLiveQuery := hrm.LastLiveQueryTime.UnixMilli()
		lastLiveQueryTime = &lastLiveQuery
	}

	jhr := JsonHintRequest{
		OrgKey:              strings.ToUpper(hrm.OrgKey), // Sanitize casing
		DeviceGroupId:       hrm.DeviceGroupId,
		DeviceId:            hrm.DeviceId,
		Version:             &hrm.Version,
		LastUpdateTime:      lastUpdate,
		LastLiveQueryTime:   lastLiveQueryTime,
		RemoteRulesRequired: hrm.RemoteRulesRequired,
		DeploymentType:      string(hrm.DeploymentType),
		DeviceType:          string(hrm.DeviceType),
	}

	if auditHintsFromCBD != nil {
		jhr.AuditHintsFromCBD = *auditHintsFromCBD
	}
	return &jhr
}

func (hrm *HintRequestMessage) MarshalJSON() ([]byte, error) {
	jhr := hrm.ToJsonMessage()
	return json.Marshal(jhr)
}
