// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package hints

import (
	"encoding/json"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/utils"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"
)

type HintResult struct {
	Hints             *hinttype.HintSet `json:"hints"`
	LastLiveQueryTime *time.Time        `json:"last_live_query_time"`
	LastUpdateTime    *time.Time        `json:"last_update_time"`
}

type jsonHintResult struct {
	Hints             *hinttype.HintSet `json:"hints"`
	LastLiveQueryTime *int64            `json:"last_live_query_time,omitempty"`
	LastUpdateTime    *int64            `json:"last_update_time,omitempty"`
}

// MarshalJSON -- Produce a JSON representation from an existing HintResult
func (hintResult *HintResult) MarshalJSON() ([]byte, error) {
	jhr := jsonHintResult{}

	if hintResult.Hints != nil {
		jhr.Hints = hintResult.Hints
	} else {
		jhr.Hints = hinttype.NewHintSet()
	}

	if hintResult.LastLiveQueryTime != nil {
		lqTime := hintResult.LastLiveQueryTime.UnixMilli()
		jhr.LastLiveQueryTime = &lqTime
	}

	if hintResult.LastUpdateTime != nil {
		updateTime := hintResult.LastUpdateTime.UnixMilli()
		jhr.LastUpdateTime = &updateTime
	}

	return json.Marshal(&jhr)
}

// UnmarshalJSON -- Overlay a JSON representation onto an existing HintResult structure
func (hintResult *HintResult) UnmarshalJSON(b []byte) error {
	jhr := new(jsonHintResult)
	err := json.Unmarshal(b, jhr)
	if err != nil {
		return err
	}

	hintResult.LastLiveQueryTime = utils.GetTimePointerFromPointer(jhr.LastLiveQueryTime)
	hintResult.LastUpdateTime = utils.GetTimePointerFromPointer(jhr.LastUpdateTime)
	hintResult.Hints = jhr.Hints
	return nil
}

func (hintResult *HintResult) addHint(ht hinttype.HintType) {
	hintResult.Hints.SetHint(ht)
}

////////////////////////////////////////////////////////////////////////////////
// Methods used for unit tests
//

func (hintResult *HintResult) AnyHintsSet() bool {
	return hintResult.numHintsSet() > 0
}

func (hintResult *HintResult) IsHintSet(ht hinttype.HintType) bool {
	return hintResult.Hints.IsHintSet(ht)
}

func (hintResult *HintResult) numHintsSet() int {
	return hintResult.Hints.Count()
}

func NewHintResult(hints ...hinttype.HintType) *HintResult {
	return &HintResult{Hints: hinttype.NewHintSet(hints...), LastLiveQueryTime: nil, LastUpdateTime: nil}
}
